package gearman

import (
	"context"
	"log"
	"sync"
	"time"

	"github.com/it-novum/gearman-go/client"
	"github.com/it-novum/gearman-go/worker"
)

// ConnectionStatus is an enum holding the current status of the connection
type ConnectionStatus int

const (
	// Connected is returend if the client has an active connection to the server
	Connected ConnectionStatus = 0
	// Disconnected is returend if the client can't connect to the server
	Disconnected ConnectionStatus = 1
	// Stopped is returend if the client has been closed
	Stopped ConnectionStatus = 2
)

type ClientType int

const (
	Main      ClientType = 0
	Satellite ClientType = 1
)

type ClientOptions struct {
	ClientType     ClientType
	ClientName     string
	Address        string
	MessageReceive chan<- *Message
}

// Client represents the connection to gearman server
type Client struct {
	ClientOptions

	mtx      sync.Mutex
	wg       sync.WaitGroup
	shutdown chan struct{}

	worker *worker.Worker
	client *client.Client

	messageSend *SyncMessageFIFO

	statusClient          ConnectionStatus
	statusWorker          ConnectionStatus
	workerDisconnectError *worker.WorkerDisconnectError
}

// IsRunning returns true if gearman worker and client is running
func (c *Client) IsRunning() bool {
	c.mtx.Lock()
	defer c.mtx.Unlock()

	if c.worker != nil && c.client != nil && c.statusClient == Connected && c.statusWorker == Connected {
		return true
	}
	return false
}

func (c *Client) connectWorker() {
	c.mtx.Lock()
	defer c.mtx.Unlock()

	if c.worker == nil {
		log.Println(&Error{clientName: c.ClientName, msg: "connect to gearman server as worker"})
		w := worker.New(worker.OneByOne)

		if err := w.AddServer("tcp", c.Address); err != nil {
			log.Println(&Error{clientName: c.ClientName, err: err})
			return
		}

		if c.ClientType == Main {
			for _, messageType := range MessageTypeMain {
				messageType := messageType // copy by value
				if err := w.AddFunc(messageType.WorkerQueue(), func(job worker.Job) ([]byte, error) {
					// Add gearman message into FIFO
					c.MessageReceive <- &Message{Type: messageType, String: string(job.Data())}
					return nil, nil
				}, worker.Unlimited); err != nil {
					log.Println(&Error{clientName: c.ClientName, err: err})
					return
				}
			}
		} else {
			for _, messageType := range MessageTypeSatellite {
				messageType := messageType // copy by value
				if err := w.AddFunc(messageType.WorkerQueue(), func(job worker.Job) ([]byte, error) {
					// Add gearman message into FIFO
					c.MessageReceive <- &Message{Type: messageType, String: string(job.Data())}
					return nil, nil
				}, worker.Unlimited); err != nil {
					log.Println(&Error{clientName: c.ClientName, err: err})
					return
				}
			}
		}

		w.ErrorHandler = func(err error) {
			c.mtx.Lock()
			defer c.mtx.Unlock()

			// if it is already disconnected we get into a reconnect loop
			if c.statusWorker == Connected {
				// Store the Reconnect func for later reconnect
				disconnect, ok := err.(*worker.WorkerDisconnectError)
				if ok {
					c.statusWorker = Disconnected
					c.workerDisconnectError = disconnect
				}
			}
		}

		// w.Ready() is going to connect to the gearman server
		if err := w.Ready(); err != nil {
			log.Println(&Error{clientName: c.ClientName, err: err, msg: "connection failed"})
			return
		}
		log.Println(&Error{clientName: c.ClientName, msg: "connection to gearman successful"})

		go w.Work()
		c.worker = w
		c.statusWorker = Connected
	} else if c.statusWorker == Disconnected {
		log.Println(&Error{clientName: c.ClientName, msg: "reconnect required"})

		// Do not run the Reconnect func inside of the error handler !
		c.workerDisconnectError.Reconnect()
		c.statusWorker = Connected
		c.workerDisconnectError = nil
	}
}

// Set client status to Disconnected in case of any error
func (c *Client) handleClientError(err error) {
	c.mtx.Lock()
	defer c.mtx.Unlock()

	if c.statusClient == Connected {
		log.Println(&Error{clientName: c.ClientName, err: err, msg: "connection failed"})
		c.statusClient = Disconnected
		if c.client != nil {
			c.client.Close()
		}
		c.client = nil
	}
}

// Create new GearmanClient to put messages into a gearman queue
func (c *Client) connectClient() {
	c.mtx.Lock()
	defer c.mtx.Unlock()

	if c.client == nil {
		log.Println(&Error{clientName: c.ClientName, msg: "connect to gearman server as client"})
		gc, err := client.New("tcp", c.Address)
		if err != nil {
			log.Println(&Error{clientName: c.ClientName, err: err, msg: "connection failed"})
			return
		}

		gc.ErrorHandler = c.handleClientError

		c.client = gc
		c.statusClient = Connected
	} else if c.statusClient == Disconnected {
		c.statusClient = Connected
	}
}

// run the event loop in a gropher func in background
func (c *Client) run(parent context.Context) {
	ctx, cancel := context.WithCancel(parent)
	defer cancel()

	defer c.wg.Done()

	defer func() {
		if c.worker != nil {
			log.Println(&Error{clientName: c.ClientName, msg: "close gearman worker connection"})
			c.worker.Close()
			c.worker = nil
		}
		if c.client != nil {
			log.Println(&Error{clientName: c.ClientName, msg: "close gearman client connection"})
			_ = c.client.Close()
			c.client = nil
		}
		log.Println(&Error{clientName: c.ClientName, msg: "close gearman fifo"})
		c.messageSend.Close()
		c.messageSend = nil
		c.statusWorker = Stopped
		c.statusClient = Stopped
	}()

	// Check all 5 seconds if connection is established and reconnect if necessary
	ticker := time.NewTicker(time.Second * 5)
	defer ticker.Stop()

	// Init connect on startup
	c.connectWorker()
	c.connectClient()

	noNotify := make(chan struct{})

	for {
		var notify chan struct{}
		if c.IsRunning() {
			// Only listen to notify signal from FIFO if gearman connection is established
			notify = c.messageSend.Notify()
		} else {
			log.Println(&Error{clientName: c.ClientName, msg: "Waiting for gearman reconnect"})
			notify = noNotify
		}

		select {
		case <-ctx.Done():
			return
		case _, more := <-c.shutdown:
			if !more {
				// Channel c.shutdown got closed - drop gearman client
				return
			}
		case <-ticker.C:
			c.connectWorker()
			c.connectClient()
		case <-notify:
			// retest again if we are still running
			// Got notify signal from FIFO send message to local or remote gearman
			if c.IsRunning() {
				msgs := c.messageSend.GetAll(16 * 1024)
				for i, msg := range msgs {
					if err := c.sendMessage(msg); err != nil {
						c.handleClientError(err)
						c.messageSend.PushAll(msgs[i:], true)
						break
					}
				}
			}
		}
	}
}

// Start the event loop and returns if successful
func (c *Client) Start(ctx context.Context) {
	c.wg.Add(1)
	c.shutdown = make(chan struct{})
	c.messageSend = &SyncMessageFIFO{}
	c.messageSend.Start()
	c.statusClient = Disconnected
	c.statusWorker = Disconnected

	go c.run(ctx)
}

// Stop should be called before programs exit
func (c *Client) Stop() {
	close(c.shutdown)
	c.wg.Wait()
}

func (c *Client) sendMessage(msg *Message) error {
	return c.sendRawMessage(msg.Type.ClientQueue(), msg.String)
}

func (c *Client) sendRawMessage(queue, data string) error {
	_, err := c.client.DoBg(queue, []byte(data), client.JobNormal)
	if err != nil {
		return &Error{clientName: c.ClientName, err: err, msg: "could not send gearman worker job to server"}
	}
	return nil
}

// SendMessage queues a message for sending to gearman
func (c *Client) SendMessage(msg *Message) {
	c.messageSend.Push(msg)
}
